import os
import random
import StringIO
import time

from pvHelpers.utils import CaseInsensitiveDict, randUnicode
from werkzeug.datastructures import FileStorage


def get_members(org_id, crypto_client):
    existing_users = CaseInsensitiveDict(
        {u['user_id']: u for u in crypto_client.list_local_users()['users']})
    return CaseInsensitiveDict({
        uid: u for uid, u in existing_users.iteritems() if
             u['org_info'] and u['org_info']['org_id'] == org_id
    })


def send_email(sender, recipients, crypto_client):
    subject = randUnicode()
    text = randUnicode()
    tos = recipients
    ccs = []
    bccs = []
    html = u'<div>{}</div>'.format(randUnicode())
    atts = [{
        'content': os.urandom(1024),
        'name': randUnicode(),
        'type': 'image/jpeg'
    }, {
        'content': os.urandom(1024),
        'name': randUnicode(),
        'type': 'application/zip'
    }, {
        'content': os.urandom(1024),
        'name': randUnicode(),
        'type': 'audio/jpeg'
    }, {
        'content': os.urandom(1024),
        'name': randUnicode(),
        'type': 'application/pdf'
    }]

    attachments = [
        FileStorage(
            stream=StringIO.StringIO(a['content']),
            filename=a['name'], content_type=a['type']) for a in atts]

    crypto_client.send_email(
        4, {'user_id': sender['user_id'], 'display_name': sender['display_name']},
        [{'user_id': t['user_id'], 'display_name': t['display_name']} for t in tos],
        [{'user_id': c['user_id'], 'display_name': c['display_name']} for c in ccs],
        [{'user_id': b['user_id'], 'display_name': b['display_name']} for b in bccs],
        subject, text, html, attachments, None, [], [], [])


def create_export_group(admin, members, optionals_required, crypto_client):
    group = crypto_client.create_org_approval_group(
        admin['user_id'], admin['org_info']['org_id'], 'export_group', members, optionals_required)

    crypto_client.set_roled_approval_group(
        admin['user_id'], admin['org_info']['org_id'], group['id'], group['version'], 'export_approval_group')

    return group


def add_member(admin, crypto_client, cs_client, member_id=None, department=None, role='standard'):
    member_id = member_id or u'{}@preveil.test'.format(randUnicode())
    department = department or randUnicode()

    crypto_client.invite_org_member(
        admin['user_id'], admin['org_info']['org_id'], member_id, randUnicode(), department, role)
    d = cs_client.get_user_secret(member_id, u'account_setup')
    member = crypto_client.claim_account(member_id, d['secret'], d['metadata']['key_version'])
    if role == 'admin':
        crypto_client.change_member_role(admin['user_id'], admin['org_info']['org_id'], member_id, role)
        member['org_info']['role'] = u'admin'

    return member


def create_new_org(admin_id, crypto_client):
    admin = crypto_client.create_test_account(admin_id, u'journal')
    crypto_client.create_organization(admin_id)['org_id']
    existing_users = CaseInsensitiveDict(
        {u['user_id']: u for u in crypto_client.update_and_list_local_users(admin_id)['users']})
    admin = existing_users[admin_id]
    return admin


class Timings:
    NEW_EMAIL = (os.environ.get('NEW_EMAIL', 10), 0)
    NEW_MEMBER = (os.environ.get('NEW_MEMBER', 100), 0)

    # EXPORT_GROUP_CHANGE
    # MEMBER_REKEY


def generate_load(admin_id, backend_client, crypto_client):
    existing_users = CaseInsensitiveDict(
        {u['user_id']: u for u in crypto_client.update_and_list_local_users(admin_id)['users']})
    admin = existing_users[admin_id]
    org_id = admin['org_info']['org_id']
    while True:
        current_time = time.time()

        # take a random member

        if current_time - Timings.NEW_MEMBER[1] > Timings.NEW_MEMBER[0]:
            add_member(admin, crypto_client, backend_client)

            Timings.NEW_MEMBER = (Timings.NEW_MEMBER[0], time.time())

        if current_time - Timings.NEW_EMAIL[1] > Timings.NEW_EMAIL[0]:
            members = get_members(org_id, crypto_client).values()
            sender = random.choice(members)
            recipients = random.sample(members, random.randint(1, len(members)))

            send_email(sender, recipients, crypto_client)

            Timings.NEW_EMAIL = (Timings.NEW_EMAIL[0], time.time())

        time.sleep(0.1)



from mail_relay.store.buckets import KeyBucket, TempDeviceKeyBycket
from pvHelpers.store import GetDBSession


def write_key(user_id, account_version, key, bucket, store_path):
    with GetDBSession(store_path) as session:
        try:
            session.begin()
            res = session.execute(
                '''
                INSERT INTO Store (bucket, bucket_version, key, value)
                VALUES (:bucket, :version, :key, :value)
                ''', {
                    'bucket': bucket.NAME,
                    'version': bucket.CURRENT_VERSION,
                    'key': bucket.key(user_id, account_version, key.key_version),
                    'value': bucket.serialize(key, bucket.CURRENT_VERSION)
                }
            )
            session.commit()
            return res.lastrowid
        except Exception:
            session.rollback()
            raise


def write_user_key(user_id, account_version, user_key, store_path):
    return write_key(
        user_id, account_version, user_key, KeyBucket, store_path)


def write_temp_device_key(user_id, account_version, device_key, store_path):
    return write_key(
        user_id, account_version, device_key, TempDeviceKeyBycket, store_path)


def read_key(user_id, account_version, key_version, bucket, store_path):
    with GetDBSession(store_path) as session:
        result = session.execute(
            '''
            SELECT id, value, bucket_version FROM Store
            WHERE bucket = :bucket AND key = :key
            ''', {
                'bucket': bucket.NAME,
                'key': bucket.key(user_id, account_version, key_version)
            }
        )
        (_, blob, b_v) = result.fetchall()[0]

    return bucket.deserialize(blob, b_v)


def read_user_key(user_id, account_version, key_version, store_path):
    return read_key(
        user_id, account_version, key_version, KeyBucket, store_path)


def read_temp_device_key(user_id, account_version, store_path):
    return read_key(
        user_id, account_version, -5, TempDeviceKeyBycket, store_path)


def read_keys(user_id, account_version, bucket, store_path):
    with GetDBSession(store_path) as session:
        result = session.execute(
            '''
            SELECT id, value, bucket_version FROM Store
            WHERE bucket = :bucket AND key LIKE :key
            ''', {
                'bucket': bucket.NAME,
                'key': bucket.key(user_id, account_version, '%')
            }
        )
        results = result.fetchall()

    return [bucket.deserialize(blob, b_v) for (_, blob, b_v) in results]


def read_user_keys(user_id, account_version, store_path):
    return read_keys(
        user_id, account_version, KeyBucket, store_path)



from mail_relay.mail.mail_state import MailState
from pvHelpers.utils import jdumps, jloads


class MailStateBucket:
    NAME = u'mail_state'
    CURRENT_VERSION = 0

    @staticmethod
    def key(user_id, account_version):
        return '{};{}'.format(account_version, user_id.lower())

    @staticmethod
    def serialize(mail_state, bucket_version):
        if bucket_version == 0:
            return MailStateBucket._serialize_v0(mail_state)
        else:
            raise ValueError(u'unsupported bucket bucket_version {}'.format(bucket_version))

    @staticmethod
    def deserialize(blob, bucket_version):
        if bucket_version == 0:
            return MailStateBucket._deserialize_v0(blob)
        else:
            raise ValueError(u'unsupported bucket bucket_version {}'.format(bucket_version))

    @staticmethod
    def _serialize_v0(mail_state):
        return jdumps(mail_state.rev_id)

    @staticmethod
    def _deserialize_v0(blob):
        j = jloads(blob)
        return MailState(j)



from .state import State
from .state_machine import StateMachine


def main(store_path, for_members=[]):
    '''Start the relay daemon'''

    state = State(store_path)
    state_machine = StateMachine(state)

    state_machine.run_forever()



import os
import random

from mail_relay.config import Config, VolumeConfig
from mail_relay.daemon.state import State
from mail_relay.mail.mail_state import MailState
from mail_relay.store.handlers import (update_config, update_user_mail_state,
                                       write_exporter, write_user,
                                       write_user_key)
from pvHelpers.utils import randUnicode
import pytest

from ..utils import create_test_account, test_local_user


@pytest.mark.integration
def key_rotatation_test(cs_client, integration_config, store):
    # needs exporter
    with pytest.raises(KeyError):
        State(store)

    user = create_test_account(cs_client)

    write_user_key(user.user_id, user.account_version, user.user_key, store)
    write_user(user, store)

    # needs exporter
    with pytest.raises(KeyError):
        State(store)

    write_exporter(user.user_id, user.account_version, store)

    # needs config
    with pytest.raises(AttributeError):
        State(store)

    update_config(integration_config, store)
    state = State(store)
    assert len(state.local_users) == 2
    assert state.local_users[(user.account_version, user.user_id)].device.key == user.device.key

    state.rotate_device_key(user.user_id, user.account_version)

    assert state.local_users[(user.account_version, user.user_id)].device.key.key_version == \
        user.device.key.key_version + 1
    assert state.local_users[(user.account_version, user.user_id)].user_key == user.user_key

    # TODO: add test for intermittend rotation and handling of `401`


def relay_mime_test(get_state, tmpdir_factory):
    user = test_local_user()
    state = get_state(user)
    rev_id = random.randint(0, 999999999)
    update_user_mail_state(
        user.user_id, user.account_version, MailState(rev_id), state.store_path)

    for r in [rev_id, rev_id - 1]:
        err = state.relay_mime(user.user_id, user.account_version, randUnicode(), r, randUnicode())
        assert isinstance(err, ValueError)

    server_id = randUnicode()
    mime = randUnicode()
    export_path = os.path.abspath(str(tmpdir_factory.mktemp('export')))
    print
    state.config = Config(state.config.cs, volume=VolumeConfig(export_path))

    err = state.relay_mime(user.user_id, user.account_version, server_id, rev_id + 1, mime)
    assert err is None
    assert state.get_members_last_rev_id(user.user_id, user.account_version).rev_id == rev_id + 1

    # check volume
    user_dir = os.path.join(export_path, user.user_id.lower())
    filename = os.path.join(
        os.path.join(export_path, user.user_id.lower()),
        server_id.lower()) + '.eml'

    assert os.path.isdir(user_dir)
    with open(filename) as f:
        assert f.read() == mime



from pvHelpers.store import GetDBSession


def remove_records(record_ids, store_path):
    with GetDBSession(store_path) as session:
        try:
            session.begin()
            session.connection().execute(
                '''
                DELETE FROM Store
                WHERE id IN ({})
                '''.format(', '.join('?' for _ in record_ids)),
                record_ids
            )
            session.commit()
        except Exception:
            session.rollback()
            raise



import os
import shutil
import subprocess
import time
import random

import click
from pvHelpers.utils import jloads, MergeDicts, utf8Decode

from ..utils import crypto_send_user_key, get_rand_port

TEST_RUNNER_ROOT = os.path.dirname(os.path.abspath(__file__))
TEST_DIR = os.path.join(TEST_RUNNER_ROOT, '.test_data')
TEST_PREFIX = 'test_'
CONFIGS_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'configs')


def test_dir(test_id):
    return os.path.join(TEST_DIR, '{}{}'.format(TEST_PREFIX, test_id))


def get_test_ids():
    return [int(d.split('_')[1]) for d in os.listdir(TEST_DIR) if d.startswith(TEST_PREFIX)]


def new_test_context():
    existing_test_ids = get_test_ids()
    next_id = max([-1] + existing_test_ids) + 1
    next_dir = test_dir(next_id)
    os.mkdir(next_dir)
    # create volumes
    for v in ['crypto', 'export', 'relay']:
        os.mkdir(os.path.join(next_dir, v))

    # set config file for relay tool
    shutil.copyfile(
        os.path.join(CONFIGS_DIR, 'volume.test.yml'),
        os.path.join(next_dir, 'relay', 'config.yml'))

    return next_id


@click.command('test')
@click.option('-l', '--list-tests', 'list_tests', flag_value=True, default=False)
@click.option('-c', '--clear', 'clear', flag_value=True, default=False)
@click.option(
    '-i', '--test-id', 'test_id',
    type=click.INT, default=None,
    help='Continue specified test number.')
@click.pass_context
def cli(gctx, test_id, list_tests, clear):
    '''handle integration test runner state'''
    p_env = os.environ.copy()
    if (clear):
        for i in get_test_ids():
            subprocess.check_call(
                ['docker-compose', 'down', '-v'], cwd=TEST_RUNNER_ROOT,
                env=MergeDicts(p_env, {'COMPOSE_PROJECT_NAME': str(i), 'CRYPTO_PORT': str(0)}))

            shutil.rmtree(test_dir(i))

        gctx.exit(0)

    if (list_tests):
        test_ids = get_test_ids()
        click.echo(', '.join([str(i) for i in sorted(test_ids)]))
        gctx.exit(0)

    if test_id is None:
        # create new test context
        click.echo('creating new test context!')
        test_id = new_test_context()
        click.echo('test context created!')

        click.echo('provisioning collection server!')
        click.echo('provisioning crypto server!')
        click.echo('provisioning load generator!')
        crypto_port = get_rand_port()
        sp_env = MergeDicts(p_env, {
            'COMPOSE_PROJECT_NAME': str(test_id),
            'CRYPTO_PORT': str(crypto_port)
        })
        # provision collection server, crypto_server
        subprocess.check_call(
            ['docker-compose', 'up', '-d', '--build', '--force-recreate'],
            cwd=TEST_RUNNER_ROOT, env=sp_env)
        # idle some time for services to become reachable
        time.sleep(10)
        click.echo('crypto server instance ready!')
        click.echo('collection server instance ready!')

        # create new user/org
        click.echo('creating test organization!')
        p = subprocess.Popen(
            ['docker-compose', 'run', '--rm', 'load_generator', '-c'],
            stdout=subprocess.PIPE, cwd=TEST_RUNNER_ROOT, env=sp_env)

        assert p.wait() == 0
        organization_info = jloads(utf8Decode(p.stdout.read()))
        click.echo('created test organization!')
        click.echo(organization_info)

        # configure relay tool
        click.echo('configuring relay tool!')
        # migrate db
        subprocess.check_call(
            ['docker-compose', 'run', '--rm', 'relay', 'migrate'],
            cwd=TEST_RUNNER_ROOT, env=sp_env)
        click.echo('relay database migrated!')
        # point to config file
        subprocess.check_call(
            ['docker-compose', 'run', '--rm', 'relay', 'config', '-c', './data/config.yml'],
            cwd=TEST_RUNNER_ROOT, env=sp_env)
        click.echo('relay configuration saved!')

        # configure exporter
        p = subprocess.Popen(
            ['docker-compose', 'run', '--rm', 'relay', 'config', 'exporter', organization_info['admin_id']],
            stdout=subprocess.PIPE, cwd=TEST_RUNNER_ROOT, env=sp_env)
        send_pin_handle = crypto_send_user_key('127.0.0.1', crypto_port, organization_info['admin_id'])
        time.sleep(3)
        p.stdout.readline() # skip just info line
        pin = p.stdout.readline().split('pin: ')[1].strip()
        send_pin_handle(pin)
        assert p.wait() == 0
        click.echo('relay exporter {} configured!'.format(organization_info['admin_id']))

        # configure approvers
        for a in random.sample(organization_info['approvers'], organization_info['optionals_required']):
            p = subprocess.Popen(
                ['docker-compose', 'run', '--rm', 'relay', 'config', 'approver', a['user_id']],
                stdout=subprocess.PIPE, cwd=TEST_RUNNER_ROOT, env=sp_env)
            send_pin_handle = crypto_send_user_key('127.0.0.1', crypto_port, a['user_id'])
            time.sleep(3)
            p.stdout.readline() # skip just info line
            pin = p.stdout.readline().split('pin: ')[1].strip()
            send_pin_handle(pin)
            assert p.wait() == 0
            click.echo('relay approver {} configured!'.format(a['user_id']))

    else:
        if not os.path.isdir(test_dir(test_id)):
            click.echo('Invalid test {}'.format(test_id))
            gctx.exit(1)
        # continue test

        p_env.update(COMPOSE_PROJECT_NAME=str(test_id))

        # provision collection server, crypto_server
        subprocess.check_call(
            ['docker-compose', 'up', '-d'], cwd=TEST_RUNNER_ROOT, env=p_env)
        time.sleep(4)
        # create new user/org
        p = subprocess.Popen(
            ['docker-compose', 'run', '--rm', 'load_generator', '-c'],
            stdout=subprocess.PIPE, cwd=TEST_RUNNER_ROOT, env=p_env)

        exit_code = p.wait()
        out = p.stdout.read()
        print out
        print exit_code
        # subprocess.check_call(
        #     ['docker-compose', 'up', '-d'], cwd=TEST_RUNNER_ROOT, env=p_env)



cli()



import os

import gevent
import pvHelpers.approval_group.export_shard as ExportShard
from pvHelpers.crypto import PVKeyFactory
from pvHelpers.cs_client.utils import ExpiredDeviceKey
from pvHelpers.logger import g_log
from pvHelpers.request import UserRequest
from pvHelpers.user import OrganizationInfo, User
from pvHelpers.utils import CaseInsensitiveDict, CaseInsensitiveSet, quiet_mkdir
import requests


def deserialize_user(u):
    organiztion_info = OrganizationInfo(
        u['entity_id'], u'', u['entity_metadata']['department'],
        u['entity_metadata']['role']) if u['entity_id'] else None

    public_user_keys = []
    if u['public_key']:
        public_user_keys = [
            PVKeyFactory.deserializePublicUserKey(u['public_key'])]

    user = User(
        u['user_id'], u['account_version'], u['display_name'],
        u['mail_collection_id'], public_user_keys, organiztion_info)

    return user


def gthread_value(g):
    if g.successful():
        return g.value, None
    else:
        return None, g.exception


def export_groups_equal(e1, e2):
    e1_set = CaseInsensitiveSet([(a['account_version'], a['user_id']) for a in e1['approvers']])
    e2_set = CaseInsensitiveSet([(a['account_version'], a['user_id']) for a in e1['approvers']])
    return e1['optionals_required'] == e2['optionals_required'] and e1_set == e2_set


# TODO: check if export requests has all the required members in its payload
def export_has_all_members_keys(export_request, members):
    return False


def write_mime_to_disk(path, member_id, server_id, mime):
    try:
        member_dir = os.path.join(path, member_id.lower())
        quiet_mkdir(member_dir)

        filename = os.path.join(member_dir, server_id.lower()) + '.eml'
        # assertion for duplicates
        assert not os.path.isfile(filename)

        with open(filename, 'w') as f:
            f.write(mime)
    except Exception as e:
        return e


def fetch_approvers_export_shards(approvers, request_id, cs_client):
    threads = []
    for a in approvers:
        threads.append(
            gevent.spawn(cs_client.getEDiscoveryShards, a, a.org_info.org_id, request_id))

    gevent.joinall(threads)

    approver_shards = []
    for (approver, (shards, err)) in zip(approvers, map(lambda t: gthread_value(t), threads)):
        if err is not None:
            if isinstance(err, ExpiredDeviceKey):
                raise err
            g_log.exception(err)

            continue

        approver_shards.append((approver, shards))

    return approver_shards


def decrypt_export_shards(export_shards, members):
    decrypted_shards = CaseInsensitiveDict()  # (key_version, member_id) => [shard*]
    for approver, approver_shards in export_shards:
        for member_id, member_key_shards in approver_shards.iteritems():
            if len(member_key_shards) > 0:
                for key_shard in member_key_shards:
                    # NOTE: these fields are missing in collection server's response
                    key_shard['user_id'] = approver.user_id
                    key_shard['required'] = False
                    # collection server just renames `secret` field to `shard`!
                    key_shard['secret'] = key_shard['shard']
                    key_version = key_shard['wrapped_key_version']

                    encrypted_shard = ExportShard.fromDict(key_shard)
                    decrypted_shard = encrypted_shard.decryptShard(
                        approver.getKeyWithVersion(encrypted_shard.sharee_key_version),
                        members[member_id].getPublicUserKeyWithVersion(encrypted_shard.sharer_key_version)
                    )

                    decrypted_shards[(key_version, member_id)] = \
                        decrypted_shards.get((key_version, member_id), []) + [decrypted_shard]

    return decrypted_shards


def get_required_pks_for_decryption(export_shards):
    members_required_pks = []
    for (_, approver_shards) in export_shards:
        members_required_pks += reduce(
            lambda acc, (member_id, key_shards):
                acc + map(lambda x: (member_id.lower(), x['sharder_key_version']), key_shards),
            approver_shards.iteritems(), []
        )

    return members_required_pks


def approve_and_sign_request(approvers, request_id, cs_client):
    def _approve(approver):
        requests = cs_client.getUserApprovals(approver, u'pending', u'pending', True, 0, 10)
        for r in requests['approvals']:
            if r['request_id'] == request_id:
                cs_client.respondToOrgApproval(
                    approver, approver.org_info.org_id,
                    UserRequest(r['payload'], r['signature'], request_id), True, {})

    threads = []
    for a in approvers:
        threads.append(gevent.spawn(_approve, a))

    gevent.joinall(threads)

    for (approver, (res, err)) in zip(approvers, map(lambda t: gthread_value(t), threads)):
        if err is not None:
            if isinstance(err, ExpiredDeviceKey):
                raise err
            # accept re-approval failures
            if isinstance(err, requests.exceptions.HTTPError) and err.response.status_code == 409:
                continue
            g_log.exception(err)


def validate_approvers(approvers, cs_client):
    threads = []
    for a in approvers:
        threads.append(
            gevent.spawn(cs_client.getUserDevices, a))

    gevent.joinall(threads)

    valid_approvers = []
    for (approver, (res, err)) in zip(approvers, map(lambda t: gthread_value(t), threads)):
        if err is not None:
            if isinstance(err, ExpiredDeviceKey):
                raise err
            g_log.exception(err)
            continue

        valid_approvers.append(approver)

    return valid_approvers


def get_check_valid_approvers(users, export_group, cs_client):
    required_approvers = export_group['optionals_required']
    approvers = [(a['account_version'], a['user_id']) for a in export_group['approvers']]
    existing_approvers = [a for a in [users.get((av, i)) for (av, i) in approvers] if a is not None]
    valid_approvers = validate_approvers(existing_approvers, cs_client)
    if len(valid_approvers) < required_approvers:
        return None, 'missing required approvers'

    return valid_approvers, None



import calendar
import time

from flanker import mime
from pvHelpers.crypto.utils import CryptoException, HexEncode, Sha256Sum
from pvHelpers.logger import g_log
from pvHelpers.mail.email import EmailException, EmailV1, PROTOCOL_VERSION
from pvHelpers.utils import b64dec, EncodingException, jloads, MergeDicts, utf8Decode


# encryption_key needed for later
def decrypt_server_message(encryption_key, decrypt_key, message):
    p_meta = jloads(utf8Decode(decrypt_key.decrypt(b64dec(message['private_metadata']))))
    return MergeDicts(message, {
        'body': MergeDicts(
                    message['body'],
                    {'snippet': utf8Decode(decrypt_key.decrypt(b64dec(message['body']['snippet'])))}),
        'private_metadata': p_meta,
        'timestamp': calendar.timegm(time.strptime(message['timestamp'], '%Y-%m-%dT%H:%M:%S')),
        'attachments': map(lambda att: MergeDicts(
            att, {'name': utf8Decode(decrypt_key.decrypt(b64dec(att['name'])))}), message['attachments'])
    })


def reconstruct_content_blocks(block_ids, blocks, symm_key):
    part = ''
    print blocks
    for block_id in block_ids:
        encrypted_block = blocks.get(block_id)
        if encrypted_block is None:
            return None, KeyError('missing block with block_id {}'.format(block_id))

        encrypted_block = encrypted_block.get('data')
        if encrypted_block is None:
            print
            return None, KeyError('data missing')

        try:
            p = symm_key.decrypt(b64dec(encrypted_block))
        except (CryptoException, EncodingException) as e:
            return None, e
        part += p

    return part, None


def get_wrapped_key(server_message):
    for block in server_message['body']['blocks']:
        return block['key_version'], block['wrapped_key']
    for att in server_message['attachments']:
        for block in att['blocks']:
            return block['key_version'], block['wrapped_key']
    raise KeyError('No wrapped key provided in server message {}'.format(server_message))


def get_email_mime(email, blocks, decrypt_key):
    content_ref_ids = [
        email.body.reference_id] + map(lambda att: att.content.reference_id, email.attachments)

    contents = {}
    for reference_id in content_ref_ids:
        raw_content, err = reconstruct_content_blocks(
            reference_id.split(','), blocks, decrypt_key)
        if err is not None:
            g_log.exception(err)
            continue
        contents[reference_id] = raw_content

    if not all(map(lambda ref_id: ref_id in contents, content_ref_ids)):
        raise EmailException('Failed to reconstuct email contents, missing blocks')

    body_content = contents[email.body.reference_id]
    if email.protocol_version is PROTOCOL_VERSION.V1:
        body_mime = mime.create.from_string(contents[email.body.reference_id])
        for att in email.attachments:
            # This replacement due protocol_version 1 body serialization which uses
            # block_ids of attachments for referencing dummy nodes, while locally we use
            # attachment hash for referencing so that attachment separation can be independent
            # of any encryption or server storage upload
            att_hash = HexEncode(Sha256Sum(contents[att.content.reference_id]))
            status, body_mime = EmailV1.replaceDummyReferences(
                body_mime, {u','.join(att.content.block_ids): att_hash})
            if not status:
                raise EmailException('Failed to replace dummy references')

        body_content = body_mime.to_string()

    email.loadBody(body_content)
    for att in email.attachments:
        if att.content.reference_id in contents:
            att.loadContent(contents[att.content.reference_id])

    return email.toMime().to_string()


def wrap_for_export(mime):
    # add header
    return mime



import datetime
import time

from mail_relay.mail.mail_state import MailState
from mail_relay.store.handlers import (add_or_update_export_state, read_config,
                                       read_export_state, read_exporter, read_user_mail_state,
                                       read_users, remove_records, update_user_info,
                                       update_user_mail_state, write_temp_device_key)
from pvHelpers.approval_group import ApprovalGroup
from pvHelpers.crypto import CryptoException, PVKeyFactory
from pvHelpers.crypto.asymm_key import AsymmKeyBase
from pvHelpers.cs_client import BackendClient
from pvHelpers.logger import g_log
from pvHelpers.request import ExportRequest
from pvHelpers.utils import b64enc, CaseInsensitiveDict, utf8Encode

from .export_state import ExportState
from .helpers import (approve_and_sign_request, decrypt_export_shards, deserialize_user,
                      export_groups_equal, export_has_all_members_keys,
                      fetch_approvers_export_shards, get_check_valid_approvers,
                      get_required_pks_for_decryption, write_mime_to_disk)


ATTEMPT_RECONSTRUCTION_DELAY = 30


# handles in applications state, caching
class State(object):
    def __init__(self, store_path):
        self.store_path = store_path

        self.config = read_config(store_path)

        exporter_a_v, exporrer_id = read_exporter(store_path)
        self.exporter = self.local_users[(exporter_a_v, exporrer_id)]

        self.cs_client = BackendClient()
        self.cs_client.init(self.config.cs.http)

        self.members = CaseInsensitiveDict()
        self.export_group = None

        # (key_version, member_id) => EncryptionKey/last-check-time
        # last time check to avoid reconstruction sooner than ATTEMPT_RECONSTRUCTION_DELAY
        self._members_keys = CaseInsensitiveDict()

    @property
    def local_users(self):
        return read_users(self.store_path)

    def rotate_device_key(self, user_id, account_version):
        user = self.local_users[(account_version, user_id)]
        new_device_key = PVKeyFactory.newUserKey(user.device.key.key_version + 1)
        record_id = write_temp_device_key(
            user.user_id, user.account_version, new_device_key, self.store_path)

        self.cs_client.refreshDeviceKey(user, new_device_key)
        user.device.key = new_device_key
        update_user_info(user, self.store_path)
        remove_records([record_id], self.store_path)

    def relay_mime(self, user_id, account_version, server_id, rev_id, mime):
        if self.get_members_last_rev_id(user_id, account_version).rev_id >= rev_id:
            return ValueError(
                '{} less that users `last_rev_id` for ({}, {}).'.format(rev_id, user_id, account_version))

        err = None
        if self.config.volume:
            err = write_mime_to_disk(self.config.volume.path, user_id, server_id, mime)
        elif self.config.smtp:
            return NotImplementedError('smtp not implemented')
        else:
            return ValueError('Unknown relaying configuration')

        if err is None:
            update_user_mail_state(user_id, account_version, MailState(rev_id), self.store_path)

        return err

    def get_member(self, member_id, account_version):
        # NOTE: Unfortunately backend keeps track of export requests by key_version
        member_versions = [m for (a_v, id_), m in self.members.iteritems() if member_id.lower() == id_.lower()]
        assert len(member_versions) == 1
        return member_versions[0]

    def get_members_last_rev_id(self, user_id, account_version):
        return read_user_mail_state(user_id, account_version, self.store_path) or 0

    def get_set_member_encryption_key(self, user_id, key_version):
        k_t = self._members_keys.get((key_version, user_id))

        if k_t is None or (not isinstance(k_t, AsymmKeyBase) and (time.time() - k_t > ATTEMPT_RECONSTRUCTION_DELAY)):
            # re-fetch shards and update newly uploaded keys
            self.fetch_shards_and_update_keys()

        k_t = self._members_keys.get((key_version, user_id))
        if not isinstance(k_t, AsymmKeyBase):
            self._members_keys[(key_version, user_id)] = time.time()
            return None

        return k_t

    def get_set_export_state(self):
        (export_group, members), err = self.get_set_org_current_export_info()
        if err is not None:
            return None, err

        export_state = read_export_state(self.store_path)

        # create new export request
        if export_state is None or any([
            not export_groups_equal(export_state.group_info, export_group),
            not export_has_all_members_keys(export_state.request, members),
            export_state.valid_for < 60 * 60
        ]):
            expiry = datetime.datetime.utcnow() + datetime.timedelta(days=7)
            export_req = self.cs_client.createUserRequest(
                self.exporter, ExportRequest.new(self.exporter, {
                    'until': unicode(expiry.isoformat()),
                    'include_emails': True,
                    'include_files': False,
                    'users': map(
                        lambda u: {'user_id': u.user_id, 'key_version': u.public_user_key.key_version},
                        [m for m in members if m.isClaimed()]
                    ),
                    'from': None
                }))['request']

            export_state = ExportState(
                ExportRequest(export_req['payload'], export_req['signature'], export_req['request_id']),
                export_group, {})

            add_or_update_export_state(export_state, self.store_path)

        # set approver info if enough approvers available
        if len(export_state.approver_signatures) < export_group['optionals_required']:
            valid_approvers, err = get_check_valid_approvers(self.local_users, export_group, self.cs_client)
            if err is not None:
                return None, err

            approve_and_sign_request(valid_approvers, export_state.request.request_id, self.cs_client)

            approvers_info = CaseInsensitiveDict()
            for approver in valid_approvers:
                approvers_info[approver.user_id] = {
                    'approver_shards': {},
                    'approver_signature': b64enc(
                        approver.user_key.signing_key.sign(utf8Encode(export_state.request.signature)))
                }
            export_state = ExportState(export_state.request, export_group, approvers_info)

        return export_state, None

    def get_set_org_current_export_info(self):
        org_info = self.cs_client.getOrgInfo(
            self.exporter, self.exporter.org_info.org_id)

        if org_info['roled_approval_groups'].get('export_approval_group') is None:
            return (None, None), 'export group not configured'

        group_id = org_info['roled_approval_groups']['export_approval_group']['group_id']
        version = org_info['roled_approval_groups']['export_approval_group']['version']
        export_group = self.cs_client.getOrgAPGInfo(
            self.exporter, self.exporter.org_info.org_id, group_id, version)

        members = [deserialize_user(u) for u in org_info['users']]
        self.members = CaseInsensitiveDict({(m.account_version, m.user_id): m for m in members})
        self.export_group = export_group
        return (export_group, members), None

    def fetch_shards_and_update_keys(self):
        export_state, err = self.get_set_export_state()
        if err is not None:
            raise RuntimeError(err)

        approvers = [(a['account_version'], a['user_id']) for a in self.export_group['approvers']]
        existing_approvers = [a for a in [self.local_users.get((av, i)) for (av, i) in approvers] if a is not None]

        export_shards = fetch_approvers_export_shards(
            existing_approvers, export_state.request.request_id, self.cs_client)
        members_required_pks = get_required_pks_for_decryption(export_shards)

        # TODO: check existing and only fetch missing keys, then update `ctx.members`
        members = self.cs_client.fetchUsers(self.exporter, list(set(members_required_pks)))

        decrypted_shards = decrypt_export_shards(export_shards, members)

        # reconstruct and update state
        for (kv, member_id), available_shards in decrypted_shards.iteritems():
            curr_key = self._members_keys.get((kv, member_id))
            if curr_key is None or curr_key:
                if len(available_shards) < self.export_group['optionals_required']:
                    g_log.error('Not enough shards to reconstructing key: ({}, {})'.format(member_id, kv))
                    continue

                ag = ApprovalGroup.new(
                    member_id, [],
                    [{'user_id': u['user_id'], 'display_name': u'dummy'} for u in self.export_group['approvers']],
                    self.export_group['optionals_required'])

                try:
                    user_key = PVKeyFactory.deserializeUserKey(ag.reconstructSecret([], available_shards))
                    # this is bad user input
                    assert kv == user_key.key_version
                    self._members_keys[(kv, member_id)] = user_key.encryption_key
                except CryptoException as e:
                    g_log.error('Error reconstructing key: ({}, {})'.format(member_id, kv))
                    g_log.exception(e)
                    continue

            # TODO: check for possible bad shards or unmatching key_version , server against wrapped



from mail_relay.daemon.export_state import ExportState
from pvHelpers.request import ExportRequest
from pvHelpers.utils import jdumps, jloads


class ExportStateBucket:
    NAME = u'export_state'
    CURRENT_VERSION = 0

    @staticmethod
    def key():
        return u'export_state'

    @staticmethod
    def serialize(export_state, bucket_version):
        if bucket_version == 0:
            return ExportStateBucket._serialize_v0(export_state)
        else:
            raise ValueError(u'unsupported bucket bucket_version {}'.format(bucket_version))

    @staticmethod
    def deserialize(blob, bucket_version):
        if bucket_version == 0:
            return ExportStateBucket._deserialize_v0(blob)
        else:
            raise ValueError(u'unsupported bucket bucket_version {}'.format(bucket_version))

    @staticmethod
    def _serialize_v0(export_state):
        return jdumps({
            'request': export_state.request.toDict(),
            'approver_signatures': export_state.approver_signatures,
            'export_group': export_state.export_group
        })

    @staticmethod
    def _deserialize_v0(blob):
        j = jloads(blob)
        r = j['request']
        approvers_info = {
            id_: {'approver_shards': {}, 'approver_signature': s} for id_, s in j['approver_signatures'].iteritems()}

        return ExportState(
            ExportRequest(r['request_payload'], r['signature'], r['request_id']), j['export_group'], approvers_info)



import pprint

import click

from .context import ctx


@click.command()
@click.option('--org-info', 'show_org_info', flag_value=True, default=False)
@click.option('--export-group', 'show_export_group', flag_value=True, default=False)
@click.option('--members', 'show_members', flag_value=True, default=False)
@ctx
def info(ctx, show_org_info, show_export_group, show_members):
    '''Show information about organization, export groups and members.'''
    from mail_relay.store.handlers import read_users, read_exporter

    exporter_a_v, exporter_id = read_exporter(ctx.store.path)
    if not exporter_id:
        click.echo('No exporter configured! Try configuring exporter '
                   '```relay config exporter ${user_id}```, then try again.')
        return

    exporter = read_users(ctx.store.path)[(exporter_a_v, exporter_id)]

    org_info = ctx.cs_client.getOrgInfo(exporter, exporter.org_info.org_id)
    pp = pprint.PrettyPrinter(indent=4, )

    if (show_org_info or show_export_group or show_members) is False:  # show all if not specified
        click.echo(pp.pformat(org_info))
    else:
        if show_org_info:  # get org info
            # latest user versions
            o = dict(org_info)
            del o['users']
            click.echo(pp.pformat(o))

        # get export group info
        # get approvers/exporter info
        if show_export_group:
            if org_info['roled_approval_groups'].get('export_approval_group') is None:
                click.echo('No export group configured! Make sure to create and '
                           'set an export group for your organization and try again.')
                return
            group_id = org_info['roled_approval_groups']['export_approval_group']['group_id']
            version = org_info['roled_approval_groups']['export_approval_group']['version']
            export_group = ctx.cs_client.getOrgAPGInfo(
                exporter, exporter.org_info.org_id, group_id, version)
            click.echo(pp.pformat(export_group))

        # get members info
        if show_members:
            click.echo(pp.pformat(org_info['users']))
